#ifndef GFXVEC2_INCLUDED // -*- C++ -*-
#define GFXVEC2_INCLUDED
#if !defined(__GNUC__)
#  pragma once
#endif

/************************************************************************

  2D Vector class

  $Id: vec2.h,v 1.12 2000/12/14 17:51:49 garland Exp $

 ************************************************************************/

#include "gfx.h"

template<class T>
class TVec2 {
private:
    T elt[2];

public:
    // Standard constructors
    //
    TVec2() { *this = 0.0; }
    TVec2(T x, T y) { elt[0]=x; elt[1]=y; }
    TVec2(const TVec2<T>& v) { *this = v; }
    TVec2(T s) { *this = s; }


    // Conversion constructors
    //
    TVec2(const float  v[2]) { elt[0]=v[0]; elt[1]=v[1]; }
    TVec2(const double v[2]) { elt[0]=v[0]; elt[1]=v[1]; }


    // Access methods
    //
    operator       T*()       { return elt; }
    operator const T*() const { return elt; }

#ifndef HAVE_CASTING_LIMITS
    T& operator[](int i)       { return elt[i]; }
    T  operator[](int i) const { return elt[i]; }
    operator const T*()       { return elt; }
#endif


    // Assignment and in-place arithmetic methods
    //
    inline TVec2<T>& operator=(const TVec2<T>& v);
    inline TVec2<T>& operator=(T s);

    inline TVec2<T>& operator+=(const TVec2<T>& v);
    inline TVec2<T>& operator-=(const TVec2<T>& v);
    inline TVec2<T>& operator*=(T s);
    inline TVec2<T>& operator/=(T s);
};

////////////////////////////////////////////////////////////////////////
//
// Method definitions
//

template<class T> inline TVec2<T>& TVec2<T>::operator=(const TVec2<T>& v)
	{ elt[0]=v[0];  elt[1]=v[1];  return *this; }

template<class T> inline TVec2<T>& TVec2<T>::operator=(T s)
	{ elt[0]=elt[1]=s;  return *this; }

template<class T> inline TVec2<T>& TVec2<T>::operator+=(const TVec2<T>& v)
	{ elt[0] += v[0];   elt[1] += v[1];   return *this; }

template<class T> inline TVec2<T>& TVec2<T>::operator-=(const TVec2<T>& v)
	{ elt[0] -= v[0];   elt[1] -= v[1];   return *this; }

template<class T> inline TVec2<T>& TVec2<T>::operator*=(T s)
	{ elt[0] *= s;   elt[1] *= s;   return *this; }

template<class T> inline TVec2<T>& TVec2<T>::operator/=(T s)
	{ elt[0] /= s;   elt[1] /= s;   return *this; }

////////////////////////////////////////////////////////////////////////
//
// Operator defintions
//

template<class T>
inline TVec2<T> operator+(const TVec2<T> &u, const TVec2<T> &v)
	{ return TVec2<T>(u[0]+v[0], u[1]+v[1]); }

template<class T>
inline TVec2<T> operator-(const TVec2<T> &u, const TVec2<T> &v)
	{ return TVec2<T>(u[0]-v[0], u[1]-v[1]); }

template<class T> inline TVec2<T> operator-(const TVec2<T> &v)
	{ return TVec2<T>(-v[0], -v[1]); }

#if _MSC_VER==1200
// Normally, we use the <class T, class N> construct below to allow
// the scalar argument to be different than the template type.  This,
// for example, allows the user to write things like v/2.
// Unfortunately, Microsoft VC6.0 (aka v1200) gets confused by this.
// So we include explicit versions for the case of int's, which is by
// far the most common one.
//
  template<class T> inline TVec2<T> operator*(T s, const TVec2<T> &v)
	{ return TVec2<T>(v[0]*s, v[1]*s); }
  template<class T> inline TVec2<T> operator*(const TVec2<T> &v, T s)
	{ return s*v; }

  template<class T> inline TVec2<T> operator/(const TVec2<T> &v, T s)
	{ return TVec2<T>(v[0]/s, v[1]/s); }

  template<class T> inline TVec2<T> operator*(int s, const TVec2<T> &v)
	{ return TVec2<T>(v[0]*s, v[1]*s); }
  template<class T> inline TVec2<T> operator*(const TVec2<T> &v, int s)
	{ return s*v; }

  template<class T> inline TVec2<T> operator/(const TVec2<T> &v, int s)
	{ return TVec2<T>(v[0]/s, v[1]/s); }
#else
  template<class T, class N> inline TVec2<T> operator*(N s, const TVec2<T> &v)
	{ return TVec2<T>(v[0]*s, v[1]*s); }
  template<class T, class N> inline TVec2<T> operator*(const TVec2<T> &v, N s)
	{ return s*v; }

  template<class T, class N> inline TVec2<T> operator/(const TVec2<T> &v, N s)
	{ return TVec2<T>(v[0]/s, v[1]/s); }
#endif

template<class T> inline T operator*(const TVec2<T> &u, const TVec2<T>& v)
	{ return u[0]*v[0] + u[1]*v[1]; }

template<class T> inline TVec2<T> perp(const TVec2<T> &v)
	{ return TVec2<T>(v[1], -v[0]); }

template<class T>
inline std::ostream &operator<<(std::ostream &out, const TVec2<T> &v)
	{ return out << v[0] << " " << v[1]; }

template<class T>
inline std::istream &operator>>(std::istream &in, TVec2<T>& v)
	{ return in >> v[0] >> v[1]; }


////////////////////////////////////////////////////////////////////////
//
// Misc. function definitions
//

template<class T> inline T norm2(const TVec2<T>& v)  { return v*v; }
template<class T> inline T norm(const TVec2<T>& v)   { return sqrt(norm2(v)); }

template<class T> inline T unitize(TVec2<T>& v)
{
    T l=norm2(v);
    if( l!=1.0 && l!=0.0 )
    {
	l = sqrt(l);
	v /= l;
    }
    return l;
}

typedef TVec2<double> Vec2;
typedef TVec2<float>  Vec2f;

// GFXVEC2_INCLUDED
#endif
