#include <gfx/gfx.h>
#include <gfx/script.h>

#include <cctype>
#include <map>
#include <string>
#include <fstream>

using namespace std;

// NOTE: The command table must not be declared as a variable; it must
// be dynamically allocated with 'new'.  Because of the implicit
// registration mechanism that we use, register_command() will be
// called during the global constructor phase of the program startup.
// We cannot guarantee that the command_table constructor would be
// called before all the other ones.  So we need to take care of
// allocating it ourselves.
//
typedef map<string, script_handler> table_t;
static table_t *command_table = NULL;

void register_command(const char *name, script_handler proc)
{
    if( !command_table )
	command_table = new table_t;

    (*command_table)[name] = proc;
}

char *string_trimleft(char *line)
{
    if(!line)  return NULL;

    while( isspace(*line) )
	line++;

    return line;
}

char *string_trimright(char *line)
{
    if(!line)  return NULL;

    char *end = line + strlen(line) - 1;
    while( end>=line && isspace(*end) ) *(end--) = '\0';
    return line;
}

char *string_shift_token(char **line)
{
    char *tok = NULL;

    if( *line )
    {
        tok = *line;

        bool override = false;
        while( **line && (override || !isspace(**line)) )
        {
            if( **line == '"' ) override = !override;
            (*line)++;
        }

        if( **line )
        {
            **line = '\0';
            *line = string_trimleft(*line+1);
        }
        else
            *line = NULL;

        if( tok[0] == '"' )
        {
            tok++;
            tok[strlen(tok)-1] = '\0';
        }

    }

    return tok;
}

static bool shift_double(char **line, double *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtod(text, &end);
        if( end != text )
            return true;
    }

    return false;
}

static bool shift_float(char **line, float *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtod(text, &end);
        if( end != text )
            return true;
    }

    return false;
}


static bool shift_int(char **line, int *x)
{
    char *text = string_shift_token(line);
    if( text )
    {
        char *end;
        *x = strtol(text, &end, 0);
        if( end != text )
            return true;
    }

    return false;
}


int string_shift_numbers(char **line, double *x, int n)
{
    int got;
    for(got=0; got<n && shift_double(line, &x[got]); got++)
	;
    return got;
}

int string_shift_numbers(char **line, float *x, int n)
{
    int got;
    for(got=0; got<n && shift_float(line, &x[got]); got++)
	;
    return got;

}

int string_shift_numbers(char **line, int *x, int n)
{
    int got;
    for(got=0; got<n && shift_int(line, &x[got]); got++)
	;
    return got;

}

static int __do_string(char *line)
{
    line = string_trimleft(line);

    // Ignore comments and whitespace-only lines
    if( !line[0] || line[0]=='#' )  return 0;

    const char *op = string_shift_token(&line);

    table_t::iterator i = command_table->find(op);

    if( i == command_table->end() )
    {
	cerr << "!! Undefined function '" << op << "'" << endl;
	return SCRIPT_ERR_UNDEF;
    }
    else
    {
	script_handler proc = (*command_table)[op];
	return (*proc)(op, line);
    }
}

int script_do_string(const char *line)
{
    char *mutable_line = strdup(line);
    int rc = __do_string(mutable_line);
    free(mutable_line);
    return rc;
}

int script_do_stream(istream& in)
{
    string line;

    getline(in, line);
    while( !in.eof() )
    {
	script_do_string(line.c_str());
	getline(in, line);
    }

    return SCRIPT_OK;
}

int script_do_file(const char *filename)
{
    ifstream in(filename);
    return script_do_stream(in);
}


static int script_include(const char *op, char *argline)
{
    char *filename = string_shift_token(&argline);
    if(!filename)  return SCRIPT_ERR_SYNTAX;

    return script_do_file(filename);
}


static script_defn handlers[] =
{
    {"include", script_include},
};

REGISTER_COMMANDS(handlers);
